@@ -29,6 +29,7 @@
 
 require 'googleauth/signet'
 require 'googleauth/credentials_loader'
+require 'googleauth/json_key_reader'
 require 'jwt'
 require 'multi_json'
 require 'stringio'
@@ -46,8 +47,10 @@ module Google
     #
     # cf [Application Default Credentials](http://goo.gl/mkAHpZ)
     class ServiceAccountCredentials < Signet::OAuth2::Client
-      TOKEN_CRED_URI = 'https://www.googleapis.com/oauth2/v3/token'
+      TOKEN_CRED_URI = 'https://www.googleapis.com/oauth2/v4/token'.freeze
       extend CredentialsLoader
+      extend JsonKeyReader
+      attr_reader :project_id
 
       # Creates a ServiceAccountCredentials.
       #
@@ -56,29 +59,34 @@ module Google
       def self.make_creds(options = {})
         json_key_io, scope = options.values_at(:json_key_io, :scope)
         if json_key_io
-          private_key, client_email = read_json_key(json_key_io)
+          private_key, client_email, project_id = read_json_key(json_key_io)
         else
-          private_key = ENV[CredentialsLoader::PRIVATE_KEY_VAR]
+          private_key = unescape ENV[CredentialsLoader::PRIVATE_KEY_VAR]
           client_email = ENV[CredentialsLoader::CLIENT_EMAIL_VAR]
+          project_id = ENV[CredentialsLoader::PROJECT_ID_VAR]
         end
+        project_id ||= CredentialsLoader.load_gcloud_project_id
 
         new(token_credential_uri: TOKEN_CRED_URI,
             audience: TOKEN_CRED_URI,
             scope: scope,
             issuer: client_email,
-            signing_key: OpenSSL::PKey::RSA.new(private_key))
+            signing_key: OpenSSL::PKey::RSA.new(private_key),
+            project_id: project_id)
+          .configure_connection(options)
       end
 
-      # Reads the private key and client email fields from the service account
-      # JSON key.
-      def self.read_json_key(json_key_io)
-        json_key = MultiJson.load(json_key_io.read)
-        fail 'missing client_email' unless json_key.key?('client_email')
-        fail 'missing private_key' unless json_key.key?('private_key')
-        [json_key['private_key'], json_key['client_email']]
+      # Handles certain escape sequences that sometimes appear in input.
+      # Specifically, interprets the "\n" sequence for newline, and removes
+      # enclosing quotes.
+      def self.unescape(str)
+        str = str.gsub '\n', "\n"
+        str = str[1..-2] if str.start_with?('"') && str.end_with?('"')
+        str
       end
 
       def initialize(options = {})
+        @project_id = options[:project_id]
         super(options)
       end
 
@@ -119,10 +127,12 @@ module Google
     class ServiceAccountJwtHeaderCredentials
       JWT_AUD_URI_KEY = :jwt_aud_uri
       AUTH_METADATA_KEY = Signet::OAuth2::AUTH_METADATA_KEY
-      TOKEN_CRED_URI = 'https://www.googleapis.com/oauth2/v3/token'
-      SIGNING_ALGORITHM = 'RS256'
+      TOKEN_CRED_URI = 'https://www.googleapis.com/oauth2/v4/token'.freeze
+      SIGNING_ALGORITHM = 'RS256'.freeze
       EXPIRY = 60
       extend CredentialsLoader
+      extend JsonKeyReader
+      attr_reader :project_id
 
       # make_creds proxies the construction of a credentials instance
       #
@@ -135,29 +145,21 @@ module Google
         new(json_key_io: args[0][:json_key_io])
       end
 
-      # Reads the private key and client email fields from the service account
-      # JSON key.
-      def self.read_json_key(json_key_io)
-        json_key = MultiJson.load(json_key_io.read)
-        fail 'missing client_email' unless json_key.key?('client_email')
-        fail 'missing private_key' unless json_key.key?('private_key')
-        [json_key['private_key'], json_key['client_email']]
-      end
-
       # Initializes a ServiceAccountJwtHeaderCredentials.
       #
       # @param json_key_io [IO] an IO from which the JSON key can be read
       def initialize(options = {})
         json_key_io = options[:json_key_io]
         if json_key_io
-          private_key, client_email = self.class.read_json_key(json_key_io)
+          @private_key, @issuer, @project_id =
+            self.class.read_json_key(json_key_io)
         else
-          private_key = ENV[CredentialsLoader::PRIVATE_KEY_VAR]
-          client_email = ENV[CredentialsLoader::CLIENT_EMAIL_VAR]
+          @private_key = ENV[CredentialsLoader::PRIVATE_KEY_VAR]
+          @issuer = ENV[CredentialsLoader::CLIENT_EMAIL_VAR]
+          @project_id = ENV[CredentialsLoader::PROJECT_ID_VAR]
         end
-        @private_key = private_key
-        @issuer = client_email
-        @signing_key = OpenSSL::PKey::RSA.new(private_key)
+        @project_id ||= CredentialsLoader.load_gcloud_project_id
+        @signing_key = OpenSSL::PKey::RSA.new(@private_key)
       end
 
       # Construct a jwt token if the JWT_AUD_URI key is present in the input
